#include <stdio.h>
#include <windows.h>
#include <time.h>

#include "SynKit.h"

int connectionTypes[] = {SE_ConnectionCOM, SE_ConnectionPS2, SE_ConnectionUSB};
char* connectionTypeNames[] = {"COM", "PS2", "USB"};
int connectionTypeCount = 3;

int deviceTypes[] = {SE_DeviceMouse, SE_DeviceTouchPad, SE_DeviceWheelMouse, SE_DeviceIBMCompatibleStick,
	SE_DeviceStyk, SE_DeviceFiveButtonWheelMouse, SE_DevicecPad};
char* deviceTypeNames[] = {"Mouse", "TouchPad", "WheelMouse", "IBMCompatibleStick", "Styk", "FiveButtonWheelMouse", "cPad"};
int deviceTypeCount = 7;

int wantedDeviceTypes[] = {SE_DeviceIBMCompatibleStick, SE_DeviceStyk};
char* wantedDeviceTypeNames[] = {"IBMCompatibleStick", "Styk"};
int wantedDeviceTypeCount = 2;

void listDevices(ISynAPI *pAPI)
{
#ifdef _DEBUG
	__time64_t ltime;
	_time64( &ltime );
	printf("Device list: %s", _ctime64( &ltime ) );

	for (int conn = 0; conn < connectionTypeCount; ++conn)
		for (int dev = 0; dev < deviceTypeCount; ++dev)
		{
			long handle = -1;

			while (pAPI->FindDevice(connectionTypes[conn], deviceTypes[dev], &handle) == SYN_OK)
			{
				printf("Found: %s on %s\n", deviceTypeNames[dev], connectionTypeNames[conn]);
			}
		}
#endif
}

HANDLE hEvents[101];
ISynDevice* connectedDevices[100];
HANDLE* connectedDeviceEvents = hEvents + 1;
DWORD lastBtnState[100];
BOOL lastMiddleClicked[100];
int connectedDeviceCount = 0;

char eventName[100];

void connectDevices(ISynAPI *pAPI)
{
#ifdef _DEBUG
	__time64_t ltime;
	_time64( &ltime );
	printf("Connecting to devices: %s", _ctime64( &ltime ) );
#endif

	connectedDeviceCount = 0;
	for (int conn = 0; conn < connectionTypeCount; ++conn)
		for (int dev = 0; dev < wantedDeviceTypeCount; ++dev)
		{
			long handle = -1;

			while (pAPI->FindDevice(connectionTypes[conn], wantedDeviceTypes[dev], &handle) == SYN_OK)
			{
#ifdef _DEBUG
				printf("Connecting to: %s on %s\n", wantedDeviceTypeNames[dev], connectionTypeNames[conn]);
#endif
				if (pAPI->CreateDevice(handle, &connectedDevices[connectedDeviceCount]) != SYN_OK)
				{
#ifdef _DEBUG
					printf("Cannot obtain a device object.\n");
#else
					MessageBox(NULL, "Cannot obtain a device object.", "TP Middle", MB_ICONERROR | MB_OK);
#endif
					continue;
				}

				sprintf(eventName, "tpmiddleDevice%d_%d", GetCurrentProcessId(), connectedDeviceCount);
				connectedDeviceEvents[connectedDeviceCount] = CreateEvent(0, 0, 0, eventName);
				connectedDevices[connectedDeviceCount]->SetEventNotification(connectedDeviceEvents[connectedDeviceCount]);

				lastBtnState[connectedDeviceCount] = 0;
				lastMiddleClicked[connectedDeviceCount] = FALSE;

				++connectedDeviceCount;
			}
		}
}

void disconnectDevices(ISynAPI *pAPI)
{
#ifdef _DEBUG
	__time64_t ltime;
	_time64( &ltime );
	printf("Disconnecting devices: %s", _ctime64( &ltime ) );
#endif

	for (int i = 0; i < connectedDeviceCount; ++i)
	{
		CloseHandle(connectedDeviceEvents[i]);
		connectedDevices[i]->Release();
		connectedDevices[i] = NULL;
	}

	connectedDeviceCount = 0;
}

#ifdef _DEBUG
int main(int argc, char *argv[])
#else
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
#endif
{
	ISynAPI *pAPI = NULL;

	if (SynCreateAPI(&pAPI) != SYN_OK)
	{
#ifdef _DEBUG
		printf("Cannot obtain an API object.\n");
		getchar();
#else
		MessageBox(NULL, "Cannot obtain an API object.", "TP Middle", MB_ICONERROR | MB_OK);
#endif
		exit(-1);
	}

	sprintf(eventName, "tpmiddleApi%d", GetCurrentProcessId());
	hEvents[0] = CreateEvent(0, 0, 0, eventName);
	pAPI->SetEventNotification(hEvents[0]);
	
	listDevices(pAPI);

	connectDevices(pAPI);

	SynPacket Packet;
	while (true)
	{
		DWORD res = MsgWaitForMultipleObjects(connectedDeviceCount + 1, hEvents, FALSE, INFINITE, QS_ALLINPUT);
		if (res == WAIT_OBJECT_0)
		{
			listDevices(pAPI);
			disconnectDevices(pAPI);
			connectDevices(pAPI);
		}
		else if (res <= WAIT_OBJECT_0 + connectedDeviceCount)
		{
			int devNum = res - WAIT_OBJECT_0 - 1;
			while (connectedDevices[devNum]->LoadPacket(Packet) != SYNE_FAIL)
			{
#ifdef _DEBUG
				__time64_t ltime;
				_time64( &ltime );
				printf("Packet: device %d, state %x, xyz %d %d %d, lastM: %d, lastB: %x, %s", devNum, Packet.ButtonState(), Packet.XDelta(), 
					Packet.YDelta(), Packet.ZDelta(), lastMiddleClicked[devNum], lastBtnState[devNum], _ctime64( &ltime ) );
#endif
				if (Packet.IsMotion())
					lastMiddleClicked[devNum] = FALSE;

				if ((!(lastBtnState[devNum] & SF_ButtonMiddle)) && (Packet.ButtonState() & SF_ButtonMiddle) && !Packet.IsMotion())
					lastMiddleClicked[devNum] = TRUE;

				if (lastMiddleClicked[devNum] && !(Packet.ButtonState() & SF_ButtonMiddle))
				{
					INPUT Input={0};

					Sleep(50);
					ZeroMemory(&Input,sizeof(INPUT));
					Input.type = INPUT_MOUSE;
					Input.mi.dwFlags = MOUSEEVENTF_MIDDLEDOWN;
					SendInput(1,&Input,sizeof(INPUT));

					Sleep(50);
					ZeroMemory(&Input,sizeof(INPUT));
					Input.type = INPUT_MOUSE;
					Input.mi.dwFlags = MOUSEEVENTF_MIDDLEUP;
					SendInput(1,&Input,sizeof(INPUT));
#ifdef _DEBUG
					__time64_t ltime;
					_time64( &ltime );
					printf("Middle click: device %d, %s", devNum, _ctime64( &ltime ) );
#endif
					lastMiddleClicked[devNum] = FALSE;
				}

				lastBtnState[devNum] = Packet.ButtonState();
			}
		}
		else if (res == WAIT_OBJECT_0 + connectedDeviceCount + 1)
		{
			MSG msg;
			while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else
			return FALSE;
	}

	disconnectDevices(pAPI);
	pAPI->Release();

	return 0;
}
